/* fixedCapacityVector.h
 */
#ifndef _FIXED_CAPACITY_VECTOR_H
#define _FIXED_CAPACITY_VECTOR_H

#include "osl/misc/carray.h"
#include "osl/misc/cstdint.h"
#include <algorithm>
#include <cstddef>
#include <cassert>

#if (__GNUC__ >= 4 && __GNUC_MINOR__ >=4)
#  pragma GCC diagnostic ignored "-Wstrict-aliasing"
#endif

namespace osl
{
  namespace misc
  {
  template <typename T, size_t Capacity>
  class FixedCapacityVector
  {
  protected:
    struct Array : public CArray<T, Capacity> {}
#ifdef __GNUC__
    __attribute__((__may_alias__))
#endif
      ;
    typedef Array array_t;
    CArray<int64_t, (sizeof(T[Capacity])+sizeof(int64_t)-1)/sizeof(int64_t)> relements;
    size_t index;
  private:
    const array_t &elements() const{
      return *reinterpret_cast<const array_t*>(&relements);
    }
    array_t &elements(){
      return *reinterpret_cast<array_t*>(&relements);
    }
  public:
    typedef typename array_t::iterator iterator;
    typedef typename array_t::const_iterator const_iterator;
    
    FixedCapacityVector() : index(0) {}
    ~FixedCapacityVector() 
    {
      std::_Destroy(begin(),end());
    }
    FixedCapacityVector(FixedCapacityVector const& rhs){
      index=rhs.size();
      std::uninitialized_copy(rhs.begin(),rhs.end(),begin());
    }
    FixedCapacityVector& operator=(FixedCapacityVector const& rhs){
      if (this == &rhs)
	return *this;
      
      if(size()>rhs.size()){
	iterator it=std::copy(rhs.begin(),rhs.end(),begin());
	std::_Destroy(it,end());
      }
      else{
	iterator it=std::copy(&(rhs.elements()[0]),
			      &(rhs.elements()[size()]),begin());
	std::uninitialized_copy(&(rhs.elements()[size()]),
				&(rhs.elements()[rhs.size()]),it);
      }
      index=rhs.size();
      return *this;
    }

    T& operator[] (size_t i)
    {
      assert(i < size());
      return elements()[i];
    }

    iterator begin() {  return &elements()[0]; }
    iterator end() { return &elements()[0]+index; }

    T& front() { return *begin(); }
    T& back() { return *(end() - 1); }

    void push_back(const T& e)
    {
      assert(index < Capacity);
      std::_Construct(&elements()[index++],e);
    }
    template <class RangeIterator>
    void push_back(const RangeIterator& first, const RangeIterator& last);
    void pop_back() { 
      std::_Destroy(&elements().elements[index]); // work around for range over in &elements[index]
      --index; 
    }
    // ʬdestructorƤ
    void clear() { 
      std::_Destroy(begin(),end());
      index = 0; 
    }

    void erase(const T& e)
    {
      const iterator new_end = std::remove(begin(), end(), e);
      std::_Destroy(new_end,end());
      index -= (end() - new_end);
    }

    /** ʣǤ */
    void unique()
    {
      std::sort(begin(),end());
      iterator last = std::unique(begin(), end());
      std::_Destroy(last,end());
      index = (last - begin());
    }

    size_t size() const { return index; }
    bool empty() const { return size() == 0; }
    size_t capacity() const { return Capacity; }

    T const& operator[] (size_t i) const
    {
      assert(i < size());
      return elements()[i];
    }
    const_iterator begin() const { return &elements()[0]; }
    const_iterator end()   const { return &elements()[0]+index; }

    const T& front() const { return *begin(); }
    const T& back() const { return *(end() - 1); }
    
    bool isMember(const T& e, const_iterator first, const_iterator last) const
    {
      return std::find(first, last, e) != last;
    }
    bool isMember(const T& e) const
    {
      return isMember(e, begin(), end());
    }
  };
    template <typename T, size_t C> inline
    bool operator==(const FixedCapacityVector<T,C>& l, const FixedCapacityVector<T,C>& r) 
    {
      return l.size() == r.size() && std::equal(l.begin(), l.end(), r.begin());
    }
  } // namespace misc
  using misc::FixedCapacityVector;
} // namespace osl

template <typename T, size_t Capacity>
template <class RangeIterator>
void osl::misc::FixedCapacityVector<T,Capacity>::push_back(const RangeIterator& first, const RangeIterator& last)
{
  iterator insert_point = end();
  index += (last - first);
  assert(index <= Capacity);
  std::uninitialized_copy(first, last, insert_point);
}


#endif /* _FIXED_CAPACITY_VECTOR_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
