/* kingArea.cc
 */
#include "osl/eval/experimental/kingArea.h"
#include "osl/centering5x3.h"

int osl::eval::experimental::
KingArea::areaIdBlack(Position king, Position center, Position target)
{
  int ix = (king.x() > 5) 
    ? (target.x() - (center.x()-2)) 
    : center.x()+2-target.x();
  if (ix < 0 || ix >= 5)
    return -1;
  int iy = target.y() - (center.y()-2);
  if (iy < 0 || iy >= 4)
    return -1;
  return ix * range_y + iy;
}

int osl::eval::experimental::
KingArea::areaIdWhite(Position king, Position center, Position target)
{
  int ix = (king.x() >= 5) 
    ? (target.x() - (center.x()-2)) 
    : center.x()+2-target.x();
  if (ix < 0 || ix >= 5)
    return -1;
  int iy = center.y()+2 - target.y();
  if (iy < 0 || iy >= 4)
    return -1;
  return ix * range_y + iy;
}

int osl::eval::experimental::
KingArea::evalWithUpdate(const NumEffectState& state, Move moved, int last_value, const Weights& values)
{
  if (moved.isPass())
    return last_value;
  if (moved.ptype() == KING)
    return eval(state, values);
  
  int result = last_value;
  const bool black_moved = moved.player() == BLACK;
  {
    const Position black = state.getKingPosition(BLACK);
    const Position bcenter = Centering5x3::adjustCenter(black);

    // adjust from
    if (! moved.isDrop()) {
      int area_id = areaIdBlack(black, bcenter, moved.from());
      if (area_id >= 0) {
	result -= values.value(index(BLACK, area_id, black, black_moved, moved.oldPtype()));
	result += values.value(index(BLACK, area_id, black, false, PTYPE_EMPTY));
      }
    }

    // adjust to
    int area_id = areaIdBlack(black, bcenter, moved.to());
    if (area_id >= 0) {
      result += values.value(index(BLACK, area_id, black, black_moved, moved.ptype()));
      // adjust capture
      if (moved.capturePtype() == PTYPE_EMPTY)
	result -= values.value(index(BLACK, area_id, black, false, PTYPE_EMPTY));
      else
	result -= values.value(index(BLACK, area_id, black, !black_moved, moved.capturePtype()));
    }
  }
  const bool white_moved = !black_moved;
  {
    const Position white = state.getKingPosition(WHITE);
    const Position wcenter = Centering5x3::adjustCenter(white);

    // adjust from
    if (! moved.isDrop()) {
      int area_id = areaIdWhite(white, wcenter, moved.from());
      if (area_id >= 0) {
	result += values.value(index(WHITE, area_id, white, white_moved, moved.oldPtype()));
	result -= values.value(index(WHITE, area_id, white, false, PTYPE_EMPTY));
      }
    }

    // adjust to
    int area_id = areaIdWhite(white, wcenter, moved.to());
    if (area_id >= 0) {
      result -= values.value(index(WHITE, area_id, white, white_moved, moved.ptype()));
      // adjust capture
      if (moved.capturePtype() == PTYPE_EMPTY)
	result += values.value(index(WHITE, area_id, white, false, PTYPE_EMPTY));
      else
	result += values.value(index(WHITE, area_id, white, !white_moved, moved.capturePtype()));
    }
  }
  return result;
}

int osl::eval::experimental::
KingArea::index(Player king_owner, int ix, int iy, Position king, Piece p)
{
  return index(king_owner, ix*range_y+iy, king, p.isOnBoardByOwner(king_owner), p.ptype());
}

int osl::eval::experimental::
KingArea::index(Player king_owner, int area_id, Position king, bool same, Ptype ptype)
{
  int basic_id = (area_id * PTYPE_SIZE + ptype)*2 + same;

  int king_id;
  const int x = abs(king.x()-5); // [0,4]
  if (king.canPromote(alt(king_owner))) 
  {
    // [0,15) : 5*3
    const int y = king.positionForBlack(king_owner).y() - 7;
    king_id = x*3+y;
  } 
  else
  {
    king_id = 15+x;
  }
  return basic_id * 32 + king_id;
}

int osl::eval::experimental::
KingArea::eval(const NumEffectState& state, const Weights& values)
{
  int result = 0;
  {
    const Position black = state.getKingPosition(BLACK);
    const Position bcenter = Centering5x3::adjustCenter(black);
    for (int ix=0; ix<5; ++ix) {
      const int x = (black.x() > 5) ? bcenter.x()-2+ix : bcenter.x()+2-ix;
      for (int iy=0; iy<4; ++iy) {
	const int y=bcenter.y()-2+iy;
	const Piece p = state.getPieceAt(Position(x,y));
	const int index = KingArea::index(BLACK, ix, iy, black, p);
	result += values.value(index);
      }
    }
  }
  {
    const Position white = state.getKingPosition(WHITE);
    const Position wcenter = Centering5x3::adjustCenter(white);
    for (int ix=0; ix<5; ++ix) {
      const int x = (white.x() >= 5) ? wcenter.x()-2+ix : wcenter.x()+2-ix;
      for (int iy=0; iy<4; ++iy) {
	const int y=wcenter.y()+2-iy;
	const Piece p = state.getPieceAt(Position(x,y));
	const int index = KingArea::index(WHITE, ix, iy, white, p);
	result -= values.value(index);
      }
    }
  }
  return result;
}

int osl::eval::experimental::
KingArea::pieceValue(const NumEffectState& state, Piece target, const Weights& values)
{
  // this value ignores Piece::EMPTY etc.
  int result = 0;
  {
    const Position black = state.getKingPosition(BLACK);
    const Position bcenter = Centering5x3::adjustCenter(black);
    int area_id = areaIdBlack(black, bcenter, target.position());
    if (area_id >= 0)
      result += values.value(index(BLACK, area_id, black, target.owner() == BLACK, target.ptype()));
  }
  {
    const Position white = state.getKingPosition(WHITE);
    const Position wcenter = Centering5x3::adjustCenter(white);
    int area_id = areaIdWhite(white, wcenter, target.position());
    if (area_id >= 0)
      result -= values.value(index(WHITE, area_id, white, target.owner() == WHITE, target.ptype()));
  }
  return result;
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
