/* SparseRegressionMultiplier.cc
 */
#include "osl/stat/sparseRegressionMultiplier.h"
#include "osl/stl/vector.h"

#include <boost/scoped_array.hpp>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include <algorithm>
#include <stdexcept>
#include <cmath>
#include <cassert>
#include <iostream>

// CREATORS
osl::stat::
SparseRegressionMultiplier::SparseRegressionMultiplier(unsigned int dimA,
						       double lambda,
						       unsigned int lambda_start)
  : m_dimA(dimA), m_lambda(lambda), m_lambda_start(lambda_start), split(0)
{
}

osl::stat::
SparseRegressionMultiplier::~SparseRegressionMultiplier()
{
}

// MANIPULATORS
// ACCESSORS
void osl::stat::
SparseRegressionMultiplier::newIteration() const
{
}

void osl::stat::
SparseRegressionMultiplier::prod(const double *x, double *out) const
{
  newIteration();
  if (m_lambda) {
    for (size_t i=0; i<m_lambda_start; ++i)
      out[i] = 0;
    for (size_t i=m_lambda_start; i<m_dimA; ++i)
      out[i] = m_lambda * x[i];
  }
  else {
    std::fill(&out[0], &out[m_dimA], 0.0);
  }

  boost::scoped_array<unsigned int> indices(new unsigned int[m_dimA]);
  boost::scoped_array<double> values(new double[m_dimA]);
  unsigned int instance_id;
  int non_zeros;
  while (getVectorXWithID(instance_id, non_zeros, &indices[0], &values[0]),
	 non_zeros >= 0)
  {
    addProduct(non_zeros, &indices[0], &values[0], x, out);
  }
}

void osl::stat::
SparseRegressionMultiplier::getVectorXWithIDPar(unsigned int& instance_id,
						int& num_elements,
						unsigned int *non_zero_indices, 
						double *non_zero_values,
						int tid) const
{
  throw std::logic_error("getVectorXWithIDPar not supported");
}

int osl::stat::
SparseRegressionMultiplier::blockSize() const
{
  return 1;			// not supported
}

void osl::stat::
SparseRegressionMultiplier::getVectorXWithIDBlock(int& num_filled,
						  vector<unsigned int>& instance_id,
						  vector<int>& num_elements,
						  vector<vector<unsigned int> >& non_zero_indices, 
						  vector<vector<double> >& non_zero_values) const
{
  throw std::logic_error("getVectorXWithIDBlock not supported");
}


typedef volatile int atomic_int_t;

struct osl::stat::SparseRegressionMultiplier::
ProdDiag
{
  const SparseRegressionMultiplier *multiplier;
  boost::mutex *mutex;
  atomic_int_t *stop;
  const double *d, *x;
  double *out;
  int tid;
  ProdDiag(const SparseRegressionMultiplier *mult, boost::mutex *m, atomic_int_t *s,
           const double *id, const double *ix, double *o, int itid)
    : multiplier(mult), mutex(m), stop(s), d(id), x(ix), out(o), tid(itid)
  {
  }
  void operator()()
  {
    boost::scoped_array<unsigned int> indices(new unsigned int[multiplier->dim()]);
    boost::scoped_array<double> values(new double[multiplier->dim()]);
    unsigned int instance_id;
    int processed = 0, non_zeros;
    while (true) {
      {
	boost::mutex::scoped_lock lk(*mutex);
        if (*stop) {
	  // std::cerr << ":" << tid << *stop;
	  ++(*stop);
          break;
	}
        multiplier->getVectorXWithID(instance_id, non_zeros, &indices[0], &values[0]);
        if (non_zeros < 0) {
	  // std::cerr << "!" << tid;
          ++(*stop);
	  break;
	}
      }
      ++processed;
      multiplier->addProductDiag(non_zeros, &indices[0], &values[0], d[instance_id], x, out);
    }
  }
};

struct osl::stat::SparseRegressionMultiplier::
ProdDiagBlock
{
  const SparseRegressionMultiplier *multiplier;
  boost::mutex *mutex;
  atomic_int_t *stop;
  const double *d, *x;
  double *out;
  int tid;
  ProdDiagBlock(const SparseRegressionMultiplier *mult, boost::mutex *m, atomic_int_t *s,
		const double *id, const double *ix, double *o, int itid)
    : multiplier(mult), mutex(m), stop(s), d(id), x(ix), out(o), tid(itid)
  {
  }
  void operator()()
  {
    const int block_size = multiplier->blockSize();
    vector<vector<unsigned int> > indices(block_size);
    vector<vector<double> > values(block_size);
    
    vector<int> non_zeros(block_size);
    vector<unsigned int> instance_id(block_size);
    
    int processed = 0, num_filled;
    while (true) {
      {
	boost::mutex::scoped_lock lk(*mutex);
        if (*stop) {
	  // std::cerr << ":" << tid << *stop;
	  ++(*stop);
          break;
	}
        multiplier->getVectorXWithIDBlock(num_filled, instance_id, non_zeros, indices, values);
        if (num_filled == 0) {
	  // std::cerr << "!" << tid;
          ++(*stop);
	  break;
	}
      }
      for (int i=0; i<num_filled; ++i) {
	++processed;
	multiplier->addProductDiag(non_zeros[i], &*indices[i].begin(), &*values[i].begin(),
				   d[instance_id[i]], x, out);
      }
    }
  }
};

struct osl::stat::SparseRegressionMultiplier::
ProdDiagPar
{
  const SparseRegressionMultiplier *multiplier;
  const double *d, *x;
  double *out;
  int tid;
  ProdDiagPar(const SparseRegressionMultiplier *mult, 
	      const double *id, const double *ix, double *o, int itid)
    : multiplier(mult), d(id), x(ix), out(o), tid(itid)
  {
  }
  void operator()()
  {
    boost::scoped_array<unsigned int> indices(new unsigned int[multiplier->dim()]);
    boost::scoped_array<double> values(new double[multiplier->dim()]);
    unsigned int instance_id;
    int non_zeros, processed = 0;
    while (true) {
      multiplier->getVectorXWithIDPar(instance_id, non_zeros, &indices[0], &values[0], tid);
      if (non_zeros < 0)
	break;
      ++processed;
      multiplier->addProductDiag(non_zeros, &indices[0], &values[0], d[instance_id], x, out);
    }
  }
};

void osl::stat::
SparseRegressionMultiplier::prodDiag(const double *d, const double *x, double *out) const
{
  newIteration();
  if (m_lambda) {
    for (size_t i=0; i<m_lambda_start; ++i)
      out[i] = 0;
    for (size_t i=m_lambda_start; i<m_dimA; ++i)
      out[i] = m_lambda * x[i];
  }
  else {
    std::fill(&out[0], &out[m_dimA], 0.0);
  }

  if (split > 0) 
  {
    boost::mutex mutex;
    const bool use_one_to_one_queue = false; // true;
    atomic_int_t stop;
    stop = 0;
    boost::ptr_vector<boost::thread> threads;
    osl::vector<osl::vector<double> > work(split);
    for (int i=0; i<split; ++i)
      work[i].resize(m_dimA, 0);
    if (use_one_to_one_queue) 
    {
      for (int i=0; i<split; ++i)
	threads.push_back(new boost::thread(ProdDiagPar(this, d, x, &*work[i].begin(), i)));
    }
    else if (blockSize() > 1) 
    {
      for (int i=0; i<split; ++i)
	threads.push_back(new boost::thread(ProdDiagBlock(this, &mutex, &stop, d, x, &*work[i].begin(), i)));
    }
    else
    {
      for (int i=0; i<split; ++i)
	threads.push_back(new boost::thread(ProdDiag(this, &mutex, &stop, d, x, &*work[i].begin(), i)));
    }
    
    for (int i=0; i<split; ++i) {
      threads[i].join();
      assert(use_one_to_one_queue || stop);
      for (size_t j=0; j<m_dimA; ++j)
	out[j] += work[i][j];
    }
    return;
  }
  boost::scoped_array<unsigned int> indices(new unsigned int[m_dimA]);
  boost::scoped_array<double> values(new double[m_dimA]);
  int non_zeros;
  unsigned int instance_id;
  while (getVectorXWithID(instance_id, non_zeros, &indices[0], &values[0]),
	 non_zeros >= 0)
  {
    addProductDiag(non_zeros, &indices[0], &values[0], d[instance_id], x, out);
  }
}

void osl::stat::
SparseRegressionMultiplier::prodRaw(const double *x, double *out) const
{
  newIteration();		// umm
  boost::scoped_array<unsigned int> indices(new unsigned int[m_dimA]);
  boost::scoped_array<double> values(new double[m_dimA]);
  int non_zeros;
  unsigned int instance_id;
  while (getVectorXWithID(instance_id, non_zeros, &indices[0], &values[0]),
	 non_zeros >= 0)
  {
    out[instance_id] = 0.0;
    for (int ip=0; ip<non_zeros; ++ip) {
      const unsigned int p = indices[ip];
      assert(p < m_dimA);
      const double vp = values[ip];
      out[instance_id] += x[p]*vp;
    }
  }
}

unsigned int osl::stat::SparseRegressionMultiplier::
computeXtY(const double *y, double *xty, double *diag_inv) const
{
  boost::scoped_array<unsigned int> indices(new unsigned int[dim()]);
  boost::scoped_array<double> values(new double[dim()]);
	
  std::fill(&xty[0], &xty[dim()], 0.0);
  std::fill(&diag_inv[0], &diag_inv[dim()], m_lambda);

  unsigned int instance_id;
  int non_zeros;
  while (getVectorXWithID(instance_id, non_zeros, &indices[0], &values[0]),
	 non_zeros >= 0)
  {
    for (int ip=0; ip<non_zeros; ++ip)
    {
      const unsigned int p = indices[ip];
      const double vp = values[ip];
#if 0
      if (p >= dim())
	std::cerr << p << " >= " << dim() << "\n";
#endif
      assert(p < dim());
      xty[p] += vp * y[instance_id];
      diag_inv[p] += vp*vp;
    }
  }

  for (unsigned int i=0; i<dim(); ++i) {
    diag_inv[i] = (fabs(diag_inv[i]) > 1e-10) ? 1.0/diag_inv[i] : 1.0;	    
  }
  return instance_id;
}

unsigned int osl::stat::SparseRegressionMultiplier::
computeXtYDiag(const double *weight, const double *y, double *xty, double *diag_inv) const
{
  boost::scoped_array<unsigned int> indices(new unsigned int[dim()]);
  boost::scoped_array<double> values(new double[dim()]);
	
  std::fill(&xty[0], &xty[dim()], 0.0);
  std::fill(&diag_inv[0], &diag_inv[dim()], m_lambda);

  unsigned int instance_id;
  int non_zeros;
  while (getVectorXWithID(instance_id, non_zeros, &indices[0], &values[0]),
	 non_zeros >= 0) {
    for (int ip=0; ip<non_zeros; ++ip) {
      const unsigned int p = indices[ip];
      const double vp = values[ip];
#ifdef DEBUG_ALL
      if (p >= dim())
	std::cerr << p << " >= " << dim() << "\n";
#endif
      assert(p < dim());
#ifdef DEBUG_ALL
      double old = xty[p];
#endif
      xty[p] += vp * y[instance_id] * weight[instance_id];
#ifdef DEBUG_ALL
      if (! std::isnan(old) && std::isnan(xty[p]))
	std::cerr << " become nan " << old << " " << vp << " " << y[instance_id] << " " << weight[instance_id] << "\n";
#endif
      assert(! std::isnan(xty[p]));
      diag_inv[p] += vp*vp * weight[instance_id];
    }
  }

  for (unsigned int i=0; i<dim(); ++i) {
    diag_inv[i] = (fabs(diag_inv[i]) > 1e-10) ? 1.0/diag_inv[i] : 1.0;	    
  }
  return instance_id;
}

void osl::stat::
SparseRegressionMultiplier::addProduct(unsigned int non_zeros,
				       const unsigned int *indices, 
				       const double *values,
				       const double *x, double *out) const
{
  assert(non_zeros <= m_dimA);
  for (unsigned int ip=0; ip<non_zeros; ++ip)
  {
    const unsigned int p = indices[ip];
    assert(p < m_dimA);
    const double vp = values[ip];
    const double xpvp = x[p]*vp;
    out[p] += xpvp*vp;
    for (unsigned int iq=ip+1; iq<non_zeros; ++iq)
    {
      const unsigned int q = indices[iq];
      const double vq = values[iq];
      assert(q < m_dimA);
      out[p] += x[q]*vp*vq;
      out[q] += xpvp*vq;
    }
  }
}

void osl::stat::
SparseRegressionMultiplier::
addProductDiag(unsigned int non_zeros, const unsigned int *indices, 
	       const double *values,
	       const double d, const double *x, double *out) const
{
  assert(non_zeros <= m_dimA);
  for (unsigned int ip=0; ip<non_zeros; ++ip)
  {
    const unsigned int p = indices[ip];
    assert(p < m_dimA);
    const double vp = values[ip];
    const double xpvp = x[p]*vp;
    out[p] += d*xpvp*vp;
    for (unsigned int iq=ip+1; iq<non_zeros; ++iq)
    {
      const unsigned int q = indices[iq];
      const double vq = values[iq];
      assert(q < m_dimA);
      out[p] += d*x[q]*vp*vq;
      out[q] += d*xpvp*vq;
    }
  }
}

unsigned int osl::stat::
SparseRegressionMultiplier::dim() const
{
  return m_dimA;
}

/* ------------------------------------------------------------------------- */
// CREATORS
osl::stat::
SparseRegressionMultiplierSeq::SparseRegressionMultiplierSeq(unsigned int dimA,
							     double lambda,
							     unsigned int lambda_start)
  : SparseRegressionMultiplier(dimA, lambda, lambda_start),
    current_id(0)
{
}

osl::stat::
SparseRegressionMultiplierSeq::~SparseRegressionMultiplierSeq()
{
}

void osl::stat::
SparseRegressionMultiplierSeq::getVectorXWithID(unsigned int& instance_id,
						int& num_elements,
						unsigned int *non_zero_indices, 
						double *non_zero_values) const
{
  instance_id = current_id++;
  getVectorX(num_elements, non_zero_indices, non_zero_values);
}

/* ------------------------------------------------------------------------- */
// CREATORS
osl::stat::
BinarySparseRegressionMultiplier::
BinarySparseRegressionMultiplier(unsigned int dimA, double lambda)
  : m_dimA(dimA), m_lambda(lambda)
{
}

osl::stat::
BinarySparseRegressionMultiplier::~BinarySparseRegressionMultiplier()
{
}

// MANIPULATORS
// ACCESSORS
void osl::stat::
BinarySparseRegressionMultiplier::newIteration() const
{
}

void osl::stat::
BinarySparseRegressionMultiplier::prod(const double *x, double *out) const
{
  newIteration();
  if (m_lambda) {
    for (size_t i=0; i<m_dimA; ++i) {
      out[i] = m_lambda * x[i];
    }
  }
  else {
    std::fill(&out[0], &out[m_dimA], 0.0);
  }

  boost::scoped_array<unsigned int> indices(new unsigned int[m_dimA]);
  int non_zeros;
  while (getVectorX(non_zeros, &indices[0]),
	 non_zeros >= 0)
  {
    addProduct(non_zeros, &indices[0], x, out);
  }
}

void osl::stat::
BinarySparseRegressionMultiplier::addProduct(unsigned int non_zeros,
					     const unsigned int *indices, 
					     const double *x, double *out) const
{
  assert(non_zeros <= m_dimA);
  for (unsigned int ip=0; ip<non_zeros; ++ip)
  {
    const unsigned int p = indices[ip];
    // assert(p < m_dimA);
    out[p] += x[p];
    for (unsigned int iq=ip+1; iq<non_zeros; ++iq)
    {
      const unsigned int q = indices[iq];
      // assert(q < m_dimA);
      out[p] += x[q];
      out[q] += x[p];
    }
  }
}

unsigned int osl::stat::
BinarySparseRegressionMultiplier::dim() const
{
  return m_dimA;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
